require 'tk'

#
# ###########################################################
#
# Message Box section:
#
# NOTE: Using this for Error messages only. For some unknown reason, when using this for comparing
# files [displaying "Difference/No Difference" popup windows], the previous/current ruby text widget
# data is not visible on the screen while the message window is up, only when you click the 'ok' button, 
# [thus the window goes away], does the ruby code become visible.
#
# I needed to switch to the Tk version of messageBox in order to get the desireed result - the previous/current
# data scrolls on the screen, then if a difference, the popup window appears saying so, or pops up at the end 
# when there is no difference detected.
#

require 'Win32API'

#
# Button styles:
#
$MB_OK = 0
$MB_OKCANCEL = 1
$MB_ABORTRETRYIGNORE = 2
$MB_YESNOCANCEL = 3
$MB_YESNO = 4
$MB_RETRYCANCEL = 5

#
# Icon styles:
#
$MB_ICONHAND = 16
$MB_ICONQUESTION = 32
$MB_ICONEXCLAMATION = 48
$MB_ICONASTERISK = 64

#
# Default button:
#
$MB_DEFBUTTON1 = 0
$MB_DEFBUTTON2 = 256
$MB_DEFBUTTON3 = 512

#
# Return values:
#
$IDOK = 1
$IDCANCEL = 2
$IDABORT = 3
$IDRETRY = 4
$IDIGNORE = 5
$IDYES = 6
$IDNO = 7

def messageBox(message, title = 'Message', style = 0)
	mb = Win32API.new("user32", "MessageBox", ['L', 'P', 'P', 'L'], 'L')
	return mb.Call(0, message, title, style)
end

#
# ###########################################################
#
# Globals/Default values @ startup:
#
global_dir = "C:/Users/"
prev_dir = "C:/Users/"
curr_dir = "C:/Users/"

#
# A little overkill on the enunciation of the path and comparison settings, but this will aid in giving 
# both a textual and a checkbox checked/not checked indication. This way we do not need to keep
# clicking on the menu buttons to check the current status.
#
path_separate_var = TkVariable.new
path_separate_var.value = '1'
path_same_var = TkVariable.new
path_same_var.value = '0'
same_path_str = "Same Previous and Current File Paths"
separate_path_str = "Separate Previous and Current File Paths"

#
# Matching the default value for path_separate_var.value.
#
path_str = separate_path_str

compare_sof_var = TkVariable.new
compare_sof_var.value = '0'
compare_sorc_var = TkVariable.new
compare_sorc_var.value = '1'
compare_sof_str = "Compare From Start of File"
compare_sorc_str = "Compare From Start of Ruby Code"

#
# Matching the default value for compare_sorc_var.value.
#
compare_str = compare_sorc_str

#
# ###############################################################
#
# Methods:
#

def process_errors(error)

	err_msg = [
	"Previous file and/or directory does not exist.",
	"Current file and/or directory does not exist."
	]

	#
	# Using bitwise operations, error is set on a bit by bit basis, where each bit represents an error message.
	#
	if (error > 0)

		mask = 1

		#
		# Math.frexp(error) returns a two element array containing the normalized fraction (a Float) and exponent
		# (a Fixnum) of error. In our case, the fraction is 0.5 and the exponent is a power of 2, as our error is 1 x 2
		# raised to 0,1,2,3,... So this returns 0.5 x 2[a power]. So, say error = 4, which is 1 x 2[2], which is the 3rd
		# error in our list. .frexp returns [0.5, 3] - we want the 3 [3rd error], we use 3 - 1 since we are starting
		# from 0 not 1. Just doing this so that I do not have to keep updating the 0.upto( ) each time I add a new error.
		#
		0.upto(Math.frexp(error)[1]-1) do |shift|
			if (((mask << shift) & error) > 0)
				messageBox(err_msg[shift], "Error!", $MB_ICONEXCLAMATION|$MB_OK)
			end
		end
	end
end

def get_file_name(dir)

	ftypes = [
		["Ruby source files", '.rbw'],
		["XML files", '.xml'],
		["All files", '*']
		]

	Tk.getOpenFile(:filetypes=>ftypes, :initialdir=>dir)
end

#
# ###########################################################
#
# Window drawing section.
#
# ###########################################################
#

font_name = "arial 10"

root = TkRoot.new() { title "CODE COMPARISON TOOL [.XML or .RBW]" }
root.geometry('+5+360')

#
# ###########################################################
#
# Frames.
#

#
# This is the default at startup.
#
compare_path_label = TkLabel.new(root, :text=>"#{compare_str} - #{path_str}", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 12").pack(:fill=>:x)

previous_frame = TkFrame.new(root, :borderwidth=>3, :relief=>:ridge).pack("side"=>"left", "padx" =>10, "pady" =>20)
previous_label = TkLabel.new(previous_frame, :text=>"Previous Version:", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>:x)

prev_file_name_var = TkVariable.new

prev_file_name_entry = TkEntry.new(previous_frame, :textvariable=>prev_file_name_var, :width=>67, "font"=>"arial 10").pack(:side=>"top", :pady=>5, :padx=>5)
prev_file_name_entry.focus

prev_bar_frame = TkFrame.new(previous_frame, :borderwidth=>1, :relief=>:ridge).pack("side"=>"top")
prev_bar_vert = TkScrollbar.new(prev_bar_frame).pack("side"=>"right", "fill"=>"y")
prev_text = TkText.new(prev_bar_frame, "wrap"=>"none", "width"=>65, "height"=>40, "font"=>font_name).pack("fill"=>"x", "expand"=>true)
prev_text.yscrollbar(prev_bar_vert)
prev_bar_horiz = TkScrollbar.new(prev_bar_frame, "orient"=>"hor").pack("side"=>"bottom", "fill"=>"x")
prev_text.xscrollbar(prev_bar_horiz)

current_frame = TkFrame.new(root, :borderwidth=>3, :relief=>:ridge).pack("side"=>"left", "padx" =>10, "pady" =>20)
current_label = TkLabel.new(current_frame, :text=>"Current Version:", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:fill=>:x)

curr_file_name_var = TkVariable.new

curr_file_name_entry = TkEntry.new(current_frame, :textvariable=>curr_file_name_var, :width=>67, "font"=>"arial 10").pack(:side=>"top", :pady=>5, :padx=>5)

curr_bar_frame = TkFrame.new(current_frame, :borderwidth=>1, :relief=>:ridge).pack("side"=>"top")
curr_bar_vert = TkScrollbar.new(curr_bar_frame).pack("side"=>"right", "fill"=>"y")
curr_text = TkText.new(curr_bar_frame, "wrap"=>"none", "width"=>65, "height"=>40, "font"=>font_name).pack("fill"=>"x", "expand"=>true)
curr_text.yscrollbar(curr_bar_vert)
curr_bar_horiz = TkScrollbar.new(curr_bar_frame, "orient"=>"hor").pack("side"=>"bottom", "fill"=>"x")
curr_text.xscrollbar(curr_bar_horiz)

#
# ###########################################################
#
# Main Menu:
#
main_menu = TkMenu.new(root)

path_menu = TkMenu.new(main_menu, :tearoff=>false)

path_menu.add('checkbutton', :variable=>path_separate_var, :label=> "Separate Previous and Current File Paths", :command=> proc {

	if (path_separate_var.value == '0')

		path_same_var.value = '1'
		path_str = same_path_str
	else
		path_same_var.value = '0'
		path_str = separate_path_str
	end

	compare_path_label.configure("text", "#{compare_str} - #{path_str}")
})

path_menu.add('separator')

path_menu.add('checkbutton', :variable=>path_same_var, :label=> "Same Previous and Current File Paths", :command=> proc {

	if (path_same_var.value == '0')

		path_separate_var.value = '1'
		path_str = separate_path_str
	else
		path_separate_var.value = '0'
		path_str = same_path_str
	end

	compare_path_label.configure("text", "#{compare_str} - #{path_str}")
})

main_menu.add('cascade', :menu=>path_menu, :label=>"Paths")

main_menu.add('command', :label=> "Previous",
:command=> proc {

	fname = get_file_name((path_same_var.value == '1' ? global_dir : prev_dir))

	if (!fname.empty?)
		if (path_same_var.value == '1')
			global_dir = File.dirname(fname)
		else
			prev_dir = File.dirname(fname)
		end

		prev_file_name_var.value = fname
		prev_text.delete("1.0", "end")
	end

	prev_file_name_entry.focus
})

main_menu.add('command', :label=> "Current",
:command=> proc {

	fname = get_file_name((path_same_var.value == '1' ? global_dir : curr_dir))

	if (!fname.empty?)
		if (path_same_var.value == '1')
			global_dir = File.dirname(fname)
		else
			curr_dir = File.dirname(fname)
		end

		curr_file_name_var.value = fname
		curr_text.delete("1.0", "end")
	end

	curr_file_name_entry.focus
})

main_menu.add('command', :label=> "Compare",
:command=> proc {

	pname = prev_file_name_var.value.strip
	cname = curr_file_name_var.value.strip

	#
	# These are TK tags, used in displaying regions of text in various colors for visual impact:
	#
	# Indicating a code difference:
	#
	prev_text.tag_configure("diff", :background=>"blue", :foreground=>"white", :font=>"arial 10 bold")
	curr_text.tag_configure("diff", :background=>"blue", :foreground=>"white", :font=>"arial 10 bold")

	#
	# Normal - no difference:
	#
	prev_text.tag_configure("no_diff", :background=>"white", :foreground=>"black", :font=>"arial 10")
	curr_text.tag_configure("no_diff", :background=>"white", :foreground=>"black", :font=>"arial 10")

	prev_text.delete("1.0", "end")
	curr_text.delete("1.0", "end")

	#
	# Make sure we have something to compare:
	#
	error = 0
	error = (File.exist?(pname) ? 0 : 1) | (File.exist?(cname) ? 0 : 2)

	if (error > 0)
		process_errors(error)
	else

		prev_line = curr_line = nil

		#
		# Do we start at the beginning of the file [sof == true], or where the Ruby Code resides?
		#
		sof = compare_sof_var.value.eql?('1')
		found_prev_rc = found_curr_rc = false
		p_rc_count_down = c_rc_count_down = 4

		File.open(pname, "r") do |prev_file|

			if (sof)

				#
				# Comparing from the start of the files.
				#
				prev_line = prev_file.gets
			else

				#
				# Comparing from the start of the Ruby Code section in a form.
				#
				while (prev_line = prev_file.gets)

					case
						when (prev_line =~ /\A# \*\*\*\*\* Calculations \*\*\*\*\*\Z/): p_rc_count_down -= 1
						when (prev_line =~ /\A\$, *= *"\\x02"\Z/): p_rc_count_down -= 1
						when (prev_line =~ /\A\$\\ *= *"\\x03\\n"\Z/): p_rc_count_down -= 1
						when (p_rc_count_down == 1)
							if (!prev_line.strip.empty?)
								found_prev_rc = true
								break
							end
					end
				end
			end

			File.open(cname, "r") do |curr_file|

				if (sof)
	
					#
					# Comparing from the start of the files.
					#
					curr_line = curr_file.gets
				else
	
					#
					# Comparing from the start of the Ruby Code section in a form.
					#
					while (curr_line = curr_file.gets)

						case
							when (curr_line =~ /\A# \*\*\*\*\* Calculations \*\*\*\*\*\Z/): c_rc_count_down -= 1
							when (curr_line =~ /\A\$, *= *"\\x02"\Z/): c_rc_count_down -= 1
							when (curr_line =~ /\A\$\\ *= *"\\x03\\n"\Z/): c_rc_count_down -= 1
							when (c_rc_count_down == 1)
								if (!curr_line.strip.empty?)
									found_curr_rc = true
									break
								end
						end
					end
				end

				#
				# Read both file's a line at a time, looking for differences.
				# Each line will be output in it's own text window, the first difference
				# will be highlighted and processing will stop, with the highlighted section
				# visible.
				#
				process = true
				difference = false

				while (process)

					case
						when (!prev_line and curr_line)

							#
							# Previous file is at EOF, current file has code:
							#
							difference = true
							curr_text.insert("end", curr_line, "diff")

						when (prev_line and !curr_line)

							#
							# Previous file has code, current file is at EOF:
							#
							difference = true
							prev_text.insert("end", prev_line, "diff")

						when (!prev_line and !curr_line)

							#
							# Both files are at EOF, so we are done.
							#
							process = false

						when (prev_line.eql?(curr_line))
					
							#
							# Normal - they are equal at this point:
							#
							curr_text.insert("end", curr_line, "no_diff")
							prev_text.insert("end", prev_line, "no_diff")

						else
							#
							# Not equal:
							#
							difference = true
							curr_text.insert("end", curr_line, "diff")
							prev_text.insert("end", prev_line, "diff")
					end

					#
					# Finished comparing if a difference, or EOF for both files (process would already
					# be set to false up above in this case).
					#
					if (difference)
						process = false

					elsif (process)

						#
						# Read the files. Read again if we are at the last line of an XML file and we are
						# compaing ruby code form sections. This way, we do not get a false positive when
						# comparing a .RBW vs. an .XML file.
						#
						prev_line = prev_file.gets
						curr_line = curr_file.gets

						if (!sof)
							prev_line = prev_file.gets if ((prev_line.eql?("?>\n")) and (File.extname(pname).eql?(".xml")))
							curr_line = curr_file.gets if ((curr_line.eql?("?>\n")) and (File.extname(cname).eql?(".xml")))
						end
					end
				end

				#
				# Finished processing, popup a message box with the result: Need to use the Tk
				# messageBox instead of the winap32API version, using that version represses
				# the displaying of the curr_text/prev_text inserts until the window closes.
				#
				msg = ""

				if (!sof)

					#
					# Check if one of the files are missing their Ruby Code Section:
					#
					msg = (found_prev_rc ? "" : "*** Previous File is Lacking Ruby Code. ***\n\n")
					msg += (found_curr_rc ? "" : "*** Current File is Lacking Ruby Code. ***")
				end

				if (difference)

					prev_text.yview("end")
					curr_text.yview("end")

					msg = "A Difference Has Been Detected!\n\n#{msg}".chomp("\n\n")

					Tk.messageBox(:icon=>"error", :type=>'ok', :title=>"Difference", :message=>msg)
				else
					msg = "No Difference Has Been Detected!\n\n#{msg}".chomp("\n\n")

					Tk.messageBox(:icon=>"info", :type=>'ok', :title=>"No Difference", :message=>msg)
				end
			end
		end
	end
})

compare_menu = TkMenu.new(main_menu, :tearoff=>false)

compare_menu.add('checkbutton', :variable=>compare_sof_var, :label=> "Compare From Start of File", :command=> proc {

	if (compare_sof_var.value == '0')

		compare_sorc_var.value = '1'
		compare_str = compare_sorc_str
	else
		compare_sorc_var.value = '0'
		compare_str = compare_sof_str
	end

	compare_path_label.configure("text", "#{compare_str} - #{path_str}")
})

compare_menu.add('separator')

compare_menu.add('checkbutton', :variable=>compare_sorc_var, :label=> "Compare From Start of Ruby Code", :command=> proc {

	if (compare_sorc_var.value == '0')

		compare_sof_var.value = '1'
		compare_str = compare_sof_str
	else
		compare_sof_var.value = '0'
		compare_str = compare_sorc_str
	end

	compare_path_label.configure("text", "#{compare_str} - #{path_str}")
})

main_menu.add('cascade', :menu=>compare_menu, :label=>"Options")

main_menu.add('command', :label=> "Exit", :command=> proc{exit})

#
# Adding main_menu to the root window.
#
root.menu(main_menu)

Tk.mainloop()